#/********************************************************************************
# * Copyright (c) 2014, 2018 Cirrus Link Solutions and others
# *
# * This program and the accompanying materials are made available under the
# * terms of the Eclipse Public License 2.0 which is available at
# * http://www.eclipse.org/legal/epl-2.0.
# *
# * SPDX-License-Identifier: EPL-2.0
# *
# * Contributors:
# *   Cirrus Link Solutions - initial implementation
# ********************************************************************************/
import time
from .sparkplug_b_pb2 import Payload

seqNum = 0
bdSeq = 0

class DataSetDataType:
    Unknown = 0
    Int8 = 1
    Int16 = 2
    Int32 = 3
    Int64 = 4
    UInt8 = 5
    UInt16 = 6
    UInt32 = 7
    UInt64 = 8
    Float = 9
    Double = 10
    Boolean = 11
    String = 12
    DateTime = 13
    Text = 14

class MetricDataType:
    Unknown = 0
    Int8 = 1
    Int16 = 2
    Int32 = 3
    Int64 = 4
    UInt8 = 5
    UInt16 = 6
    UInt32 = 7
    UInt64 = 8
    Float = 9
    Double = 10
    Boolean = 11
    String = 12
    DateTime = 13
    Text = 14
    UUID = 15
    DataSet = 16
    Bytes = 17
    File = 18
    Template = 19

class ParameterDataType:
    Unknown = 0
    Int8 = 1
    Int16 = 2
    Int32 = 3
    Int64 = 4
    UInt8 = 5
    UInt16 = 6
    UInt32 = 7
    UInt64 = 8
    Float = 9
    Double = 10
    Boolean = 11
    String = 12
    DateTime = 13
    Text = 14


######################################################################
# Always request this before requesting the Node Birth Payload
######################################################################
def getNodeDeathPayload():
    payload = Payload()
    addMetric(payload, "bdSeq", None, MetricDataType.Int64, getBdSeqNum())
    return payload
######################################################################


######################################################################
# Always request this after requesting the Node Death Payload
######################################################################
def getNodeBirthPayload():
    global seqNum
    seqNum = 0
    payload = Payload()
    payload.timestamp = int(round(time.time() * 1000))
    payload.seq = getSeqNum()
    addMetric(payload, "bdSeq", None, MetricDataType.Int64, --bdSeq)
    return payload
######################################################################

######################################################################
# Get the DBIRTH payload
######################################################################
def getDeviceBirthPayload():
    payload = Payload()
    payload.timestamp = int(round(time.time() * 1000))
    payload.seq = getSeqNum()
    return payload
######################################################################


######################################################################
# Get a DDATA payload
######################################################################
def getDdataPayload():
    return getDeviceBirthPayload()
######################################################################


######################################################################
# Helper method for adding dataset metrics to a payload
######################################################################
def initDatasetMetric(payload, name, alias, columns, types):
    metric = payload.metrics.add()
    if name is not None:
        metric.name = name
    if alias is not None:
        metric.alias = alias
    metric.timestamp = int(round(time.time() * 1000))
    metric.datatype = MetricDataType.DataSet

    # Set up the dataset
    metric.dataset_value.num_of_columns = len(types)
    metric.dataset_value.columns.extend(columns)
    metric.dataset_value.types.extend(types)
    return metric.dataset_value
######################################################################


######################################################################
# Helper method for adding dataset metrics to a payload
######################################################################
def initTemplateMetric(payload, name, alias, templateRef):
    metric = payload.metrics.add()
    if name is not None:
        metric.name = name
    if alias is not None:
        metric.alias = alias
    metric.timestamp = int(round(time.time() * 1000))
    metric.datatype = MetricDataType.Template

    # Set up the template
    if templateRef is not None:
        metric.template_value.template_ref = templateRef
        metric.template_value.is_definition = False
    else:
        metric.template_value.is_definition = True

    return metric.template_value
######################################################################


######################################################################
# Helper method for adding metrics to a container which can be a
# payload or a template with a timestamp
######################################################################
#def addMetric(container, name, alias, type, value):
#    metric.timestamp = int(round(time.time() * 1000))
#    return addMetric(container, name, alias, type, value, timestamp)

######################################################################
# Helper method for adding metrics to a container which can be a
# payload or a template
######################################################################
def addMetric(container, name, alias, type, value, timestamp=int(round(time.time() * 1000))):
    metric = container.metrics.add()
    if name is not None:
        metric.name = name
    if alias is not None:
        metric.alias = alias
    metric.timestamp = timestamp

    # print( "Type: " + str(type))

    if type == MetricDataType.Int8:
        metric.datatype = MetricDataType.Int8
        if value < 0:
            value = value + 2**8
        metric.int_value = value
    elif type == MetricDataType.Int16:
        metric.datatype = MetricDataType.Int16
        if value < 0:
            value = value + 2**16
        metric.int_value = value
    elif type == MetricDataType.Int32:
        metric.datatype = MetricDataType.Int32
        if value < 0:
            value = value + 2**32
        metric.int_value = value
    elif type == MetricDataType.Int64:
        metric.datatype = MetricDataType.Int64
        if value < 0:
            value = value + 2**64
        metric.long_value = value
    elif type == MetricDataType.UInt8:
        metric.datatype = MetricDataType.UInt8
        metric.int_value = value
    elif type == MetricDataType.UInt16:
        metric.datatype = MetricDataType.UInt16
        metric.int_value = value
    elif type == MetricDataType.UInt32:
        metric.datatype = MetricDataType.UInt32
        metric.int_value = value
    elif type == MetricDataType.UInt64:
        metric.datatype = MetricDataType.UInt64
        metric.long_value = value
    elif type == MetricDataType.Float:
        metric.datatype = MetricDataType.Float
        metric.float_value = value
    elif type == MetricDataType.Double:
        metric.datatype = MetricDataType.Double
        metric.double_value = value
    elif type == MetricDataType.Boolean:
        metric.datatype = MetricDataType.Boolean
        metric.boolean_value = value
    elif type == MetricDataType.String:
        metric.datatype = MetricDataType.String
        metric.string_value = value
    elif type == MetricDataType.DateTime:
        metric.datatype = MetricDataType.DateTime
        metric.long_value = value
    elif type == MetricDataType.Text:
        metric.datatype = MetricDataType.Text
        metric.string_value = value
    elif type == MetricDataType.UUID:
        metric.datatype = MetricDataType.UUID
        metric.string_value = value
    elif type == MetricDataType.Bytes:
        metric.datatype = MetricDataType.Bytes
        metric.bytes_value = value
    elif type == MetricDataType.File:
        metric.datatype = MetricDataType.File
        metric.bytes_value = value
    elif type == MetricDataType.Template:
        metric.datatype = MetricDataType.Template
        metric.template_value = value
    else:
        print("Invalid: " + str(type))

    # Return the metric
    return metric
######################################################################

######################################################################
# Helper method for adding metrics to a container which can be a
# payload or a template
######################################################################
def addHistoricalMetric(container, name, alias, type, value):
    metric = addMetric(container, name, alias, type, value)
    metric.is_historical = True

    # Return the metric
    return metric
######################################################################

######################################################################
# Helper method for adding metrics to a container which can be a
# payload or a template
######################################################################
def addNullMetric(container, name, alias, type):
    metric = container.metrics.add()
    if name is not None:
        metric.name = name
    if alias is not None:
        metric.alias = alias
    metric.timestamp = int(round(time.time() * 1000))
    metric.is_null = True

    # print( "Type: " + str(type))

    if type == MetricDataType.Int8:
        metric.datatype = MetricDataType.Int8
    elif type == MetricDataType.Int16:
        metric.datatype = MetricDataType.Int16
    elif type == MetricDataType.Int32:
        metric.datatype = MetricDataType.Int32
    elif type == MetricDataType.Int64:
        metric.datatype = MetricDataType.Int64
    elif type == MetricDataType.UInt8:
        metric.datatype = MetricDataType.UInt8
    elif type == MetricDataType.UInt16:
        metric.datatype = MetricDataType.UInt16
    elif type == MetricDataType.UInt32:
        metric.datatype = MetricDataType.UInt32
    elif type == MetricDataType.UInt64:
        metric.datatype = MetricDataType.UInt64
    elif type == MetricDataType.Float:
        metric.datatype = MetricDataType.Float
    elif type == MetricDataType.Double:
        metric.datatype = MetricDataType.Double
    elif type == MetricDataType.Boolean:
        metric.datatype = MetricDataType.Boolean
    elif type == MetricDataType.String:
        metric.datatype = MetricDataType.String
    elif type == MetricDataType.DateTime:
        metric.datatype = MetricDataType.DateTime
    elif type == MetricDataType.Text:
        metric.datatype = MetricDataType.Text
    elif type == MetricDataType.UUID:
        metric.datatype = MetricDataType.UUID
    elif type == MetricDataType.Bytes:
        metric.datatype = MetricDataType.Bytes
    elif type == MetricDataType.File:
        metric.datatype = MetricDataType.File
    elif type == MetricDataType.Template:
        metric.datatype = MetricDataType.Template
    else:
        print( "Invalid: " + str(type))

    # Return the metric
    return metric
######################################################################


######################################################################
# Helper method for getting the next sequence number
######################################################################
def getSeqNum():
    global seqNum
    retVal = seqNum
    # print("seqNum: " + str(retVal))
    seqNum += 1
    if seqNum == 256:
        seqNum = 0
    return retVal
######################################################################


######################################################################
# Helper method for getting the next birth/death sequence number
######################################################################
def getBdSeqNum():
    global bdSeq
    retVal = bdSeq
    # print("bdSeqNum: " + str(retVal))
    bdSeq += 1
    if bdSeq == 256:
        bdSeq = 0
    return retVal
######################################################################


def addMetricDataset_from_dict(payload, name, alias, data):
    """
    Converts a dictionary into a Sparkplug B dataset metric.

    Args:
        payload (Payload): The payload to add the dataset metric to.
        name (str): The name of the dataset metric.
        alias (int): The alias for the metric.
        data (dict): The dictionary containing columns as keys and lists of values.

    Returns:
        The initialized dataset metric.

    Example:

        # Sample dictionary data
        data = {
            "Temperature": [23.5, 22.0, 21.8],
            "Humidity": [60.2, 58.9, 59.5],
            "Status": ["Normal", "Warning", "Alert"]
        }

        # Create a new Sparkplug B payload
        payload = Payload()

        # Convert the dictionary into a dataset metric
        dict_to_dataset_metric(payload, "environmental_data", 1, data)

    """
    
    # Extract columns and data types
    columns = list(data.keys())

    # Check for consistent column dimensions
    lengths = [len(v) for v in data.values()]
    if len(set(lengths)) > 1:
        raise ValueError(f"All columns must have the same number of rows, but got lengths: {lengths}")

    # Determine types based on the first row's types.
    first_row = [v[0] for v in data.values()]
    types = []
    for value in first_row:
        if isinstance(value, int):
            types.append(DataSetDataType.Int64)
        elif isinstance(value, float):
            types.append(DataSetDataType.Float)
        elif isinstance(value, str):
            types.append(DataSetDataType.String)
        else:
            raise ValueError(f"Unsupported data type: {type(value)}")

    # Initialize the dataset metric using the provided function
    dataset = initDatasetMetric(payload, name, alias, columns, types)

    # Get the number of rows from the data values
    num_rows = lengths[0]

    # Add rows to the dataset
    for i in range(num_rows):
        dataset_row = dataset.rows.add()
        for col in columns:
            element = dataset_row.elements.add()
            value = data[col][i]
            # Add the value based on its type
            if isinstance(value, int):
                element.long_value = value
            elif isinstance(value, float):
                element.float_value = value
            elif isinstance(value, str):
                element.string_value = value
            else:
                raise ValueError(f"Unsupported value type: {type(value)}")

    return dataset




import unittest
import time
from datetime import datetime
from io import BytesIO

from mqtt_spb_wrapper.spb_base import  MetricGroup, MetricValue, MetricDataType

class TestMetricGroup(unittest.TestCase):

    def test_initialization(self):
        """Test that MetricGroup initializes correctly."""
        mg = MetricGroup(birth_prefix="TEST")
        self.assertEqual(mg.birth_prefix, "TEST")
        self.assertTrue(mg.is_empty())
        self.assertEqual(mg.count(), 0)

    def test_set_value(self):
        """Test setting a single value."""
        mg = MetricGroup()
        result = mg.set_value(name="temperature", value=25.5)
        self.assertTrue(result)
        self.assertFalse(mg.is_empty())
        self.assertEqual(mg.count(), 1)
        self.assertIn("temperature", mg.get_names())
        self.assertEqual(mg.get_value("temperature"), 25.5)

    def test_set_multiple_values(self):
        """Test setting multiple values."""
        mg = MetricGroup()
        values = {"pressure": 101.3, "humidity": 60}
        mg.set_dictionary(values)
        self.assertEqual(mg.count(), 2)
        self.assertEqual(mg.get_value("pressure"), 101.3)
        self.assertEqual(mg.get_value("humidity"), 60)

    def test_update_value(self):
        """Test updating an existing value."""
        mg = MetricGroup()
        mg.set_value(name="level", value=5)
        mg.set_value(name="level", value=10)
        self.assertEqual(mg.get_value("level"), 10)

    def test_remove_value(self):
        """Test removing a value."""
        mg = MetricGroup()
        mg.set_value(name="status", value="OK")
        removed = mg.remove_value("status")
        self.assertTrue(removed)
        self.assertTrue(mg.is_empty())

    def test_remove_nonexistent_value(self):
        """Test removing a value that doesn't exist."""
        mg = MetricGroup()
        removed = mg.remove_value("nonexistent")
        self.assertFalse(removed)

    def test_is_updated(self):
        """Test the is_updated method."""
        mg = MetricGroup()
        mg.set_value(name="signal", value=1)
        self.assertTrue(mg.is_updated())
        _ = mg["signal"].value  # Access value to reset is_updated
        self.assertFalse(mg.is_updated())
        mg.set_value(name="signal", value=2)
        self.assertTrue(mg.is_updated())

    def test_clear(self):
        """Test clearing all metrics."""
        mg = MetricGroup()
        mg.set_value(name="metric1", value=100)
        mg.set_value(name="metric2", value=200)
        mg.clear()
        self.assertTrue(mg.is_empty())
        self.assertEqual(mg.count(), 0)

    def test_get_value_timestamp(self):
        """Test retrieving the timestamp of a metric."""
        mg = MetricGroup()
        timestamp = int(time.time() * 1000)
        mg.set_value(name="event", value="start", timestamp=timestamp)
        self.assertEqual(mg.get_value_timestamp("event"), timestamp)

    def test_is_list_values(self):
        """Test the is_list_values method for a metric."""
        mg = MetricGroup()
        values = [1, 2, 3]
        timestamps = [0, 0, 0]
        mg.set_value(name="readings", value=values, timestamp=timestamps)
        self.assertTrue(mg.is_list_values("readings"))
        mg.set_value(name="single_reading", value=42)
        self.assertFalse(mg.is_list_values("single_reading"))

    def test_dictionary_like_operations(self):
        """Test dictionary-like access to metrics."""
        mg = MetricGroup()
        mv = MetricValue(name="voltage", value=3.3)
        mg["voltage"] = mv
        self.assertIn("voltage", mg)
        self.assertEqual(mg["voltage"].value, 3.3)
        del mg["voltage"]
        self.assertNotIn("voltage", mg)

    def test_iteration(self):
        """Test iterating over the MetricGroup."""
        mg = MetricGroup()
        mg.set_value(name="metric1", value=10)
        mg.set_value(name="metric2", value=20)
        keys = [key for key in mg]
        self.assertListEqual(sorted(keys), ["metric1", "metric2"])

    def test_length(self):
        """Test the __len__ method."""
        mg = MetricGroup()
        self.assertEqual(len(mg), 0)
        mg.set_value(name="metric1", value=10)
        self.assertEqual(len(mg), 1)

    def test_get_dictionary(self):
        """Test getting the group's metrics as a dictionary."""
        mg = MetricGroup()
        mg.set_value(name="speed", value=88)
        mg.set_value(name="acceleration", value=9.8)
        metrics_dict = mg.get_dictionary()
        self.assertEqual(len(metrics_dict), 2)
        names = [metric['name'] for metric in metrics_dict]
        self.assertIn("speed", names)
        self.assertIn("acceleration", names)

    def test_set_value_with_callback(self):
        """Test setting a value with a callback function."""
        callback_called = False

        def callback(value):
            nonlocal callback_called
            callback_called = True
            self.assertEqual(value, 100)

        mg = MetricGroup()
        mg.set_value(name="threshold", value=50, callback_on_change=callback)
        mg.set_value(name="threshold", value=100)
        self.assertTrue(callback_called)

    def test_spb_alias_num(self):
        """Test setting and retrieving spb_alias_num."""
        mg = MetricGroup()
        mg.set_value(name="alias_metric", value=1.23, spb_alias_num=100)
        self.assertEqual(mg["alias_metric"].spb_alias_num, 100)

    def test_spb_data_type(self):
        """Test setting and retrieving spb_data_type."""
        mg = MetricGroup()
        mg.set_value(name="text_metric", value="hello", spb_data_type=MetricDataType.Text)
        self.assertEqual(mg["text_metric"].spb_data_type, MetricDataType.Text)

    def test_set_value_with_none(self):
        """Test that setting a value to None is ignored."""
        mg = MetricGroup()
        result = mg.set_value(name="null_metric", value=None)
        self.assertFalse(result)
        self.assertTrue(mg.is_empty())

    def test_set_dictionary_with_timestamps(self):
        """Test setting multiple values with a specific timestamp."""
        mg = MetricGroup()
        timestamp = int(time.time() * 1000)
        values = {"temp": 22.5, "humidity": 55}
        mg.set_dictionary(values, timestamp=timestamp)
        self.assertEqual(mg.get_value_timestamp("temp"), timestamp)
        self.assertEqual(mg.get_value_timestamp("humidity"), timestamp)

    def test_has_callbacks(self):
        """Test checking if any metrics have callbacks."""
        mg = MetricGroup()
        mg.set_value(name="metric1", value=1)
        self.assertFalse(mg["metric1"].has_callback())
        mg.set_value(name="metric2", value=2, callback_on_change=lambda x: x)
        self.assertTrue(mg["metric2"].has_callback())

    def test_is_list_values_nonexistent(self):
        """Test is_list_values method for a nonexistent metric."""
        mg = MetricGroup()
        self.assertTrue(mg.is_list_values("nonexistent"))

    def test_timestamps_mismatch(self):
        """Test handling of mismatched value and timestamp lists."""
        mg = MetricGroup()
        values = [1, 2, 3]
        timestamps = [0, 0, 0]
        mg.set_value(name="mismatch_metric", value=values, timestamp=timestamps)
        self.assertEqual(mg["mismatch_metric"].value, values)
        self.assertEqual(mg["mismatch_metric"].timestamp, timestamps)

    def test_metric_value_access(self):
        """Test accessing MetricValue instances directly."""
        mg = MetricGroup()
        mg.set_value(name="direct_metric", value=123)
        metric_value = mg["direct_metric"]
        self.assertIsInstance(metric_value, MetricValue)
        self.assertEqual(metric_value.value, 123)

    def test_metricgroup_str_repr(self):
        """Test the __str__ and __repr__ methods."""
        mg = MetricGroup()
        mg.set_value(name="test_metric", value=42)
        mg_str = str(mg)
        mg_repr = repr(mg)
        self.assertIn("test_metric", mg_str)
        self.assertIn("test_metric", mg_repr)

if __name__ == '__main__':
    unittest.main()
